#include "video_buffer.h"
#include "utils.h"
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

static compressed_images_buffer buffer;

static image_item decompressed_image;
static volatile int decompressed_image_ready;

static int image_pipe_write, image_pipe_read;

static pthread_mutex_t compressed_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t compressed_cond = PTHREAD_COND_INITIALIZER;

static pthread_mutex_t decompressed_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t decompressed_cond = PTHREAD_COND_INITIALIZER;

void put_compressed_images(char *data, int size)
{
  compressed_images_item *n;

  if (pthread_mutex_lock(&compressed_mutex)) abort();

  n = malloc(sizeof(*n)); if (n == NULL) abort();
  n->data = malloc(size); if (n->data == NULL) abort();
  memcpy(n->data, data, size);
  n->size = size;
  n->next = NULL;
  if (buffer.end != NULL) buffer.end->next = n;
  if (buffer.start == NULL) buffer.start = n;
  buffer.end = n;

  //buffer.count++;

  if (pthread_cond_broadcast(&compressed_cond)) abort();
  if (pthread_mutex_unlock(&compressed_mutex)) abort();
}

compressed_images_item get_compressed_images(void)
{
  compressed_images_item ret;
  compressed_images_item *t;

  if (pthread_mutex_lock(&compressed_mutex)) abort();
  if (its_over) goto over;
  while (buffer.start == NULL) {
    if (pthread_cond_wait(&compressed_cond, &compressed_mutex)) abort();
    if (its_over) {
over:
      ret.data = NULL; ret.size = 0;
      if (pthread_cond_broadcast(&compressed_cond)) abort();
      if (pthread_mutex_unlock(&compressed_mutex)) abort();
      return ret;
    }
  }

  ret = *buffer.start;

  t = buffer.start;
  buffer.start = t->next;
  if (t == buffer.end) buffer.end = NULL;
  free(t);

  //buffer.count--;
  //fprintf(stderr, "video buffer count %d\n", buffer.count);

  if (pthread_cond_broadcast(&compressed_cond)) abort();
  if (pthread_mutex_unlock(&compressed_mutex)) abort();

  return ret;
}

int get_image_buffer_pipe(void)
{
  return image_pipe_read;
}

void init_image_buffer(void)
{
  int p[2];
  if (pipe(p) == -1) abort();
  image_pipe_read = p[0];
  image_pipe_write = p[1];
}

void put_image(char *image, int width, int height)
{
  char dummy = 0;

  if (width <= 0 || height <= 0) abort();

  if (pthread_mutex_lock(&decompressed_mutex)) abort();
  if (its_over) goto over;
  while (decompressed_image_ready) {
    if (pthread_cond_wait(&decompressed_cond, &decompressed_mutex)) abort();
    if (its_over) {
over:
      if (pthread_cond_broadcast(&decompressed_cond)) abort();
      if (pthread_mutex_unlock(&decompressed_mutex)) abort();
      return;
    }
  }

  decompressed_image.data = malloc(width*height*4);
  if (decompressed_image.data == NULL) abort();
  memcpy(decompressed_image.data, image, width*height*4);
  decompressed_image.width = width;
  decompressed_image.height = height;

  decompressed_image_ready = 1;
  if (write(image_pipe_write, &dummy, 1) != 1) abort();

  if (pthread_cond_broadcast(&decompressed_cond)) abort();
  if (pthread_mutex_unlock(&decompressed_mutex)) abort();
}

image_item get_image(void)
{
#if 0
  struct timeval now;
  struct timespec timeout;
  int retcode;
#endif
  image_item ret;
  char dummy;

  if (pthread_mutex_lock(&decompressed_mutex)) abort();
  if (decompressed_image_ready == 0) {
    ret.data = NULL; ret.width = ret.height = 0;
    if (pthread_mutex_unlock(&decompressed_mutex)) abort();
    return ret;
  }
#if 0
  while (decompressed_image_ready == 0) {
    gettimeofday(&now, NULL);
    timeout.tv_sec = now.tv_sec;
    timeout.tv_nsec = now.tv_usec * 1000 + 100 * 1000 * 1000;
    if (timeout.tv_nsec > 1000000000) {
      timeout.tv_sec++;
      timeout.tv_nsec -= 1000000000;
    }
    retcode = pthread_cond_timedwait(&decompressed_cond,
                  &decompressed_mutex, &timeout);
    if (retcode && retcode != ETIMEDOUT) abort();
    if (retcode == ETIMEDOUT) {
      ret.data = NULL; ret.width = ret.height = 0;
      if (pthread_mutex_unlock(&decompressed_mutex)) abort();
      return ret;
    }
    if (its_over) {
      ret.data = NULL; ret.width = ret.height = 0;
      if (pthread_cond_broadcast(&decompressed_cond)) abort();
      if (pthread_mutex_unlock(&decompressed_mutex)) abort();
      return ret;
    }
  }
#endif

  ret = decompressed_image;

  decompressed_image.data = NULL;
  decompressed_image.width = 0;
  decompressed_image.height = 0;

  decompressed_image_ready = 0;
  if (read(image_pipe_read, &dummy, 1) != 1) abort();

  if (pthread_cond_broadcast(&decompressed_cond)) abort();
  if (pthread_mutex_unlock(&decompressed_mutex)) abort();

  return ret;
}

void its_over_buffer(void)
{
  if (pthread_mutex_lock(&compressed_mutex)) abort();
  if (pthread_cond_broadcast(&compressed_cond)) abort();
  if (pthread_mutex_unlock(&compressed_mutex)) abort();

  if (pthread_mutex_lock(&decompressed_mutex)) abort();
  if (pthread_cond_broadcast(&decompressed_cond)) abort();
  if (pthread_mutex_unlock(&decompressed_mutex)) abort();
}

void reset_buffer(void)
{
  char dummy;

  while (buffer.start) {
    compressed_images_item *it = buffer.start;
    free(it->data);
    buffer.start = it->next;
    free(it);
  }
  buffer.start = NULL;
  buffer.end = NULL;

  free(decompressed_image.data);
  decompressed_image.data = NULL;
  decompressed_image.width = 0;
  decompressed_image.height = 0;

  if (decompressed_image_ready)
    if (read(image_pipe_read, &dummy, 1) != 1) abort();
  decompressed_image_ready = 0;
}
